#include <ptlib.h>
#include <h323.h>
#include <h245.h>
#include "ast_h323.h"
#include "caps_h323.h"
#include "chan_h323.h"

#define DEFINE_G711_CAPABILITY(cls, code, capName) \
class cls : public AST_G711Capability { \
public: \
	cls() : AST_G711Capability(240, code) { } \
}; \
H323_REGISTER_CAPABILITY(cls, capName) \

DEFINE_G711_CAPABILITY(AST_G711ALaw64Capability, H323_G711Capability::ALaw, OPAL_G711_ALAW_64K);
DEFINE_G711_CAPABILITY(AST_G711uLaw64Capability, H323_G711Capability::muLaw, OPAL_G711_ULAW_64K);
H323_REGISTER_CAPABILITY(AST_G7231Capability, OPAL_G7231);
H323_REGISTER_CAPABILITY(AST_G729Capability,  OPAL_G729);
H323_REGISTER_CAPABILITY(AST_G729ACapability, OPAL_G729A);
H323_REGISTER_CAPABILITY(AST_GSM0610Capability, OPAL_GSM0610);
H323_REGISTER_CAPABILITY(AST_CiscoG726Capability, CISCO_G726r32);
H323_REGISTER_CAPABILITY(AST_CiscoDtmfCapability, CISCO_DTMF_RELAY);

OPAL_MEDIA_FORMAT_DECLARE(OpalG711ALaw64kFormat,
	OPAL_G711_ALAW_64K,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::PCMA,
	TRUE,	// Needs jitter
	64000,	// bits/sec
	8,		// bytes/frame
	8,		// 1 millisecond/frame
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalG711uLaw64kFormat,
	OPAL_G711_ULAW_64K,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::PCMU,
	TRUE,	// Needs jitter
	64000,	// bits/sec
	8,		// bytes/frame
	8,		// 1 millisecond/frame
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalG729Format,
	OPAL_G729,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::G729,
	TRUE,	// Needs jitter
	8000,	// bits/sec
	10,		// bytes
	80,		// 10 milliseconds
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalG729AFormat,
	OPAL_G729 "A",
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::G729,
	TRUE,	// Needs jitter
	8000,	// bits/sec
	10,		// bytes
	80,		// 10 milliseconds
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalG7231_6k3Format,
	OPAL_G7231_6k3,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::G7231,
	TRUE,	// Needs jitter
	6400,	// bits/sec
	24,		// bytes
	240,	// 30 milliseconds
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalG7231A_6k3Format,
	OPAL_G7231A_6k3,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::G7231,
	TRUE,	// Needs jitter
	6400,	// bits/sec
	24,		// bytes
	240,	// 30 milliseconds
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalGSM0610Format,
	OPAL_GSM0610,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::GSM,
	TRUE,	// Needs jitter
	13200,	// bits/sec
	33,		// bytes
	160,	// 20 milliseconds
	OpalMediaFormat::AudioTimeUnits,
	0);
OPAL_MEDIA_FORMAT_DECLARE(OpalCiscoG726Format,
	CISCO_G726r32,
	OpalMediaFormat::DefaultAudioSessionID,
	RTP_DataFrame::G726,
	TRUE,	// Needs jitter
	32000,	// bits/sec
	4,		// bytes
	8,		// 1 millisecond
	OpalMediaFormat::AudioTimeUnits,
	0);
#if 0
OPAL_MEDIA_FORMAT_DECLARE(OpalCiscoDTMFRelayFormat,
	CISCO_DTMF_RELAY,
	OpalMediaFormat::DefaultAudioSessionID,
	(RTP_DataFrame::PayloadTypes)121, // Choose this for Cisco IOS compatibility
	TRUE,	// Needs jitter
	100,	// bits/sec
	4,		// bytes/frame
	8*150,	// 150 millisecond
	OpalMediaFormat::AudioTimeUnits,
	0);
#endif

/*
 * Capability: G.711
 */
AST_G711Capability::AST_G711Capability(int frames, H323_G711Capability::Mode m, H323_G711Capability::Speed s)
	: H323AudioCapability(frames, frames)
{
	mode = m;
	speed = s;
}


PObject * AST_G711Capability::Clone() const
{
	return new AST_G711Capability(*this);
}


unsigned AST_G711Capability::GetSubType() const
{
	static const unsigned G711SubType[2][2] = {
		{ H245_AudioCapability::e_g711Alaw64k, H245_AudioCapability::e_g711Alaw56k },
		{ H245_AudioCapability::e_g711Ulaw64k, H245_AudioCapability::e_g711Ulaw56k }
	};
	return G711SubType[mode][speed];
}


PString AST_G711Capability::GetFormatName() const
{
	static const char * const G711Name[2][2] = {
		{ OPAL_G711_ALAW_64K, OPAL_G711_ALAW_56K },
		{ OPAL_G711_ULAW_64K, OPAL_G711_ULAW_56K },
	};
	return G711Name[mode][speed];
}


H323Codec * AST_G711Capability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}


/*
 * Capability: G.723.1
 */
AST_G7231Capability::AST_G7231Capability(int frames, PBoolean annexA_)
	: H323AudioCapability(frames, frames)
{
	annexA = annexA_;
}

PObject::Comparison AST_G7231Capability::Compare(const PObject & obj) const
{
	Comparison result = H323AudioCapability::Compare(obj);
	if (result != EqualTo) {
		return result;
	}
	PINDEX otherAnnexA = ((const AST_G7231Capability &)obj).annexA;
	if (annexA < otherAnnexA) {
		return LessThan;
	}
	if (annexA > otherAnnexA) {
		return GreaterThan;
	}
	return EqualTo;
}

PObject * AST_G7231Capability::Clone() const
{
	return new AST_G7231Capability(*this);
}

PString AST_G7231Capability::GetFormatName() const
{
	return (annexA ? OPAL_G7231 "A" : OPAL_G7231);
}

unsigned AST_G7231Capability::GetSubType() const
{
	return H245_AudioCapability::e_g7231;
}

PBoolean AST_G7231Capability::OnSendingPDU(H245_AudioCapability & cap,
										unsigned packetSize) const
{
	cap.SetTag(H245_AudioCapability::e_g7231);
	H245_AudioCapability_g7231 & g7231 = cap;
	g7231.m_maxAl_sduAudioFrames = packetSize;
	g7231.m_silenceSuppression = annexA;
	return TRUE;
}

PBoolean AST_G7231Capability::OnReceivedPDU(const H245_AudioCapability & cap,
										unsigned & packetSize)
{
	if (cap.GetTag() != H245_AudioCapability::e_g7231) {
		return FALSE;
	}
	const H245_AudioCapability_g7231 & g7231 = cap;
	packetSize = g7231.m_maxAl_sduAudioFrames;
	annexA = g7231.m_silenceSuppression;
	return TRUE;
}

H323Codec * AST_G7231Capability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}

/*
 * Capability: G.729
 */
AST_G729Capability::AST_G729Capability(int frames)
	: H323AudioCapability(frames, frames)
{
}

PObject * AST_G729Capability::Clone() const
{
	return new AST_G729Capability(*this);
}

unsigned AST_G729Capability::GetSubType() const
{
	return H245_AudioCapability::e_g729;
}

PString AST_G729Capability::GetFormatName() const
{
	return OPAL_G729;
}

H323Codec * AST_G729Capability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}

/*
 * Capability: G.729A
 */
AST_G729ACapability::AST_G729ACapability(int frames)
	: H323AudioCapability(frames, frames)
{
}

PObject * AST_G729ACapability::Clone() const
{
	return new AST_G729ACapability(*this);
}

unsigned AST_G729ACapability::GetSubType() const
{
	return H245_AudioCapability::e_g729AnnexA;
}

PString AST_G729ACapability::GetFormatName() const
{
	return OPAL_G729A;
}

H323Codec * AST_G729ACapability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}

/*
 * Capability: GSM full rate
 */
AST_GSM0610Capability::AST_GSM0610Capability(int frames, int comfortNoise_, int scrambled_)
	: H323AudioCapability(frames, frames)
{
	comfortNoise = comfortNoise_;
	scrambled = scrambled_;
}

PObject * AST_GSM0610Capability::Clone() const
{
	return new AST_GSM0610Capability(*this);
}

unsigned AST_GSM0610Capability::GetSubType() const
{
	return H245_AudioCapability::e_gsmFullRate;
}

PBoolean AST_GSM0610Capability::OnSendingPDU(H245_AudioCapability & cap,
										unsigned packetSize) const
{
	cap.SetTag(H245_AudioCapability::e_gsmFullRate);
	H245_GSMAudioCapability & gsm = cap;
	gsm.m_audioUnitSize = packetSize * 33;
	gsm.m_comfortNoise = comfortNoise;
	gsm.m_scrambled = scrambled;
	return TRUE;
}

PBoolean AST_GSM0610Capability::OnReceivedPDU(const H245_AudioCapability & cap,
										unsigned & packetSize)
{
	if (cap.GetTag() != H245_AudioCapability::e_gsmFullRate)
		return FALSE;
	const H245_GSMAudioCapability & gsm = cap;
	packetSize = (gsm.m_audioUnitSize + 32) / 33;
	comfortNoise = gsm.m_comfortNoise;
	scrambled = gsm.m_scrambled;

	return TRUE;
}

PString AST_GSM0610Capability::GetFormatName() const
{
	return OPAL_GSM0610;
}

H323Codec * AST_GSM0610Capability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}

/*
 * Capability: G.726 32 Kbps
 */
AST_CiscoG726Capability::AST_CiscoG726Capability(int rx_frames)
	: H323NonStandardAudioCapability(rx_frames, 240,
		181, 0, 18,
		(const BYTE *)"G726r32", 0)
{
}

PObject *AST_CiscoG726Capability::Clone() const
{
	return new AST_CiscoG726Capability(*this);
}

H323Codec *AST_CiscoG726Capability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}

PString AST_CiscoG726Capability::GetFormatName() const
{
	return PString(CISCO_G726r32);
}

/*
 * Capability: Cisco RTP DTMF Relay
 */
AST_CiscoDtmfCapability::AST_CiscoDtmfCapability()
	: H323NonStandardDataCapability(0, 181, 0, 18, (const BYTE *)"RtpDtmfRelay", 0)
{
	rtpPayloadType = (RTP_DataFrame::PayloadTypes)121;
}

PObject *AST_CiscoDtmfCapability::Clone() const
{
	return new AST_CiscoDtmfCapability(*this);
}

H323Codec *AST_CiscoDtmfCapability::CreateCodec(H323Codec::Direction direction) const
{
	return NULL;
}

PString AST_CiscoDtmfCapability::GetFormatName() const
{
	return PString(CISCO_DTMF_RELAY);
}

/* Asterisk T.38 Capability */
AST_T38Capability::AST_T38Capability(const struct t38_profile *profile) : H323_T38Capability(H323_T38Capability::e_UDP)
{
	SetProfile(profile);
}

PObject *AST_T38Capability::Clone() const
{
	return new AST_T38Capability(*this);
}

H323Channel *AST_T38Capability::CreateChannel(H323Connection &connection,
				              H323Channel::Directions direction, unsigned sessionID,
            				      const H245_H2250LogicalChannelParameters *param) const
{
	return new MyT38Channel(connection, *this, direction, sessionID);
}

void AST_T38Capability::SetProfile(const struct t38_profile* profile)
{
	m_profile.IncludeOptionalField(H245_T38FaxProfile::e_t38FaxUdpOptions);
	m_profile.m_fillBitRemoval = profile->ast.fill_bit_removal;
	m_profile.m_transcodingJBIG = profile->ast.transcoding_jbig;
	m_profile.m_transcodingMMR = profile->ast.transcoding_mmr;
	m_profile.m_version = profile->ast.version;

	switch (profile->ast.rate_management) {
	case AST_T38_RATE_MANAGEMENT_LOCAL_TCF:
		m_profile.m_t38FaxRateManagement.SetTag(H245_T38FaxRateManagement::e_localTCF);
		break;
	case AST_T38_RATE_MANAGEMENT_TRANSFERRED_TCF:
		m_profile.m_t38FaxRateManagement.SetTag(H245_T38FaxRateManagement::e_transferredTCF);
		break;
	default:
		break;	
	}

	if (profile->max_datagram) {
	        m_profile.m_t38FaxUdpOptions.IncludeOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxDatagram);
		m_profile.m_t38FaxUdpOptions.m_t38FaxMaxDatagram = profile->max_datagram;
	} else
		m_profile.m_t38FaxUdpOptions.RemoveOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxDatagram);

	switch (profile->error_correction_scheme) {
	case UDPTL_ERROR_CORRECTION_REDUNDANCY:
		m_profile.m_t38FaxUdpOptions.m_t38FaxUdpEC.SetTag(H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPRedundancy);
		break;
	case UDPTL_ERROR_CORRECTION_FEC:
		m_profile.m_t38FaxUdpOptions.m_t38FaxUdpEC.SetTag(H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPFEC);
		break;
	default:
		break;	
	}

	switch (profile->ast.rate) {
	case AST_T38_RATE_2400:
		m_bit_rate = 24;
		break;
	case AST_T38_RATE_4800:
		m_bit_rate = 48;
		break;
        case AST_T38_RATE_7200:
		m_bit_rate = 72;
		break;
	case AST_T38_RATE_9600:
		m_bit_rate = 96;
		break;
        case AST_T38_RATE_12000:
		m_bit_rate = 120;
		break;
        case AST_T38_RATE_14400:
		m_bit_rate = 144;
		break;
	default:
		break;
	}
}

void AST_T38Capability::GetProfile(struct t38_profile* profile) const
{
	memset(profile, 0, sizeof(*profile));
	
	profile->ast.fill_bit_removal = m_profile.m_fillBitRemoval;
	profile->ast.transcoding_jbig = m_profile.m_transcodingJBIG;
	profile->ast.transcoding_mmr = m_profile.m_transcodingMMR;
	profile->ast.version = m_profile.m_version;

	switch (m_profile.m_t38FaxRateManagement.GetTag()) {
	case H245_T38FaxRateManagement::e_localTCF:
		profile->ast.rate_management = AST_T38_RATE_MANAGEMENT_LOCAL_TCF;
		break;
	case H245_T38FaxRateManagement::e_transferredTCF:
		profile->ast.rate_management = AST_T38_RATE_MANAGEMENT_TRANSFERRED_TCF;
		break;
	default:
		break;	
	}

	if (m_profile.m_t38FaxUdpOptions.HasOptionalField(H245_T38FaxUdpOptions::e_t38FaxMaxDatagram))
		profile->max_datagram = m_profile.m_t38FaxUdpOptions.m_t38FaxMaxDatagram;

	switch (m_profile.m_t38FaxUdpOptions.m_t38FaxUdpEC.GetTag()) {
	case H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPRedundancy:
		profile->error_correction_scheme = UDPTL_ERROR_CORRECTION_REDUNDANCY;
		break;
	case H245_T38FaxUdpOptions_t38FaxUdpEC::e_t38UDPFEC:
		profile->error_correction_scheme = UDPTL_ERROR_CORRECTION_FEC;
		break;
	default:
		break;	
	}

	switch (m_bit_rate) {
	case 24:
		profile->ast.rate = AST_T38_RATE_2400;
		break;
	case 48:
		profile->ast.rate = AST_T38_RATE_4800;
		break;
        case 72:
		profile->ast.rate = AST_T38_RATE_7200;
		break;
	case 96:
		profile->ast.rate = AST_T38_RATE_9600;
		break;
        case 120:
		profile->ast.rate = AST_T38_RATE_12000;
		break;
        case 144:
		profile->ast.rate = AST_T38_RATE_14400;
		break;
	default:
		break;
	}
}

BOOL AST_T38Capability::OnSendingPDU(H245_DataApplicationCapability & pdu) const
{
	PTRACE(3, "ASTH323T38\tOnSendingPDU for capability");

	pdu.m_maxBitRate = m_bit_rate;
	pdu.m_application.SetTag(H245_DataApplicationCapability_application::e_t38fax);
	H245_DataApplicationCapability_application_t38fax & fax = pdu.m_application;
	fax.m_t38FaxProtocol.SetTag(H245_DataProtocolCapability::e_udp);
	fax.m_t38FaxProfile = m_profile;

	return TRUE;
}

BOOL AST_T38Capability::OnSendingPDU(H245_DataMode & pdu) const
{
	PTRACE(3, "ASTH323T38\tOnSendingPDU for capability");

	pdu.m_bitRate = m_bit_rate;
	pdu.m_application.SetTag(H245_DataMode_application::e_t38fax);
	H245_DataMode_application_t38fax & fax = pdu.m_application;
	fax.m_t38FaxProtocol.SetTag(H245_DataProtocolCapability::e_udp);
	fax.m_t38FaxProfile = m_profile;

	return TRUE;
}

BOOL AST_T38Capability::OnReceivedPDU(const H245_DataApplicationCapability & pdu)
{
        if (pdu.m_application.GetTag() != H245_DataApplicationCapability_application::e_t38fax)
		return FALSE;

	const H245_DataApplicationCapability_application_t38fax & fax = pdu.m_application;
	const H245_DataProtocolCapability & proto = fax.m_t38FaxProtocol;

	if (proto.GetTag() != H245_DataProtocolCapability::e_udp)
		return FALSE;

	m_bit_rate = pdu.m_maxBitRate;
	m_profile = fax.m_t38FaxProfile;

	return TRUE;
}

BOOL AST_T38Capability::OnReceivedPDU(const H245_DataMode & pdu)
{
        if (pdu.m_application.GetTag() != H245_DataApplicationCapability_application::e_t38fax)
		return FALSE;

	const H245_DataMode_application_t38fax & fax = pdu.m_application;
	const H245_DataProtocolCapability & proto = fax.m_t38FaxProtocol;

	if (proto.GetTag() != H245_DataProtocolCapability::e_udp)
		return FALSE;

	m_bit_rate = pdu.m_bitRate;
	m_profile = fax.m_t38FaxProfile;

	return TRUE;
}
